from fnmatch import fnmatch


def separate_input_and_chip_infiles(bed_files, input_pattern="input"):
    """Split list of files depending on whether name contains input_pattern.

    Since docopt cannot take two lists of variadic arguments, splitting it
    based on a pattern is a way to work around this.

    Keyword Arguments:
    input_pattern -- pattern to split on
    bed_files     -- list of files to split
    """

    chip_files, input_files = [], []
    input_pattern = "*" + input_pattern + "*"
    input_pattern = input_pattern.upper()

    for bed_file in bed_files:
        if fnmatch(bed_file.upper(), input_pattern):
            input_files.append(bed_file)
        else:
            chip_files.append(bed_file)

    if len(chip_files) == 0:
        raise ValueError(
            "All files considered input. This means all your files match the"
            " case insensitive pattern: " + input_pattern)
    if len(input_files) == 0:
        raise ValueError(
            "No input files found. This means none of your files match"
            " the case insensitive pattern: " + input_pattern)

    return chip_files, input_files



import logging
from functools import partial

import pandas as pd

from joblib import Parallel, delayed


def find_islands(dfs, score_threshold, args):
    logging.info("Merging bins into islands.")
    parallel_find_islands = partial(_find_islands, args.window_size,
                                    args.gaps_allowed, score_threshold)
    clustered_islands = Parallel(n_jobs=args.number_cores)(
        delayed(parallel_find_islands)(df) for df in dfs)
    return clustered_islands


def _find_islands(window_size, gaps_allowed, score_threshold, df):

    if df.empty:
        return df

    # adding one to allowed gaps, because if one bin starts at x and another at x + window_size
    # there is no distance between them
    distance_allowed = window_size * (gaps_allowed + 1)
    chromosome = df.iloc[0].Chromosome

    # Code below could probably be optimized....
    # Idea: use one function to do everything
    island_starts = df.groupby(((df.Bin - df.Bin.shift()).abs(
    ) > distance_allowed).cumsum()).apply(lambda island: island.Bin.min())
    island_ends = df.groupby(((df.Bin - df.Bin.shift()).abs(
    ) > distance_allowed).cumsum()).apply(
        lambda island: island.Bin.max() + (window_size - 1))
    island_chip_counts = df.groupby(((df.Bin - df.Bin.shift()).abs(
    ) > distance_allowed).cumsum()).apply(lambda island: island.ChIP.sum())
    island_input_counts = df.groupby(((df.Bin - df.Bin.shift()).abs(
    ) > distance_allowed).cumsum()).apply(lambda island: island.Input.sum())
    island_scores = df.groupby(((df.Bin - df.Bin.shift()).abs(
    ) > distance_allowed).cumsum()).apply(lambda island: island.Score.sum())

    df = pd.concat(
        [island_starts, island_ends, island_chip_counts, island_input_counts,
         island_scores],
        axis=1).reset_index(drop=True)

    df.columns = ["Start", "End", "ChIP", "Input", "Score"]
    df.insert(0, "Chromosome", chromosome)
    df = df[df["Score"] > score_threshold]

    return df.reset_index(drop=True)



import pytest

from collections import namedtuple

import pandas as pd

__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"

MockNamespace = namedtuple(
    "MockNamespace",
    ["number_cores", "genome", "keep_duplicates", "window_size",
     "fragment_size", "paired_end", "gaps_allowed",
     "false_discovery_rate_cutoff", "effective_genome_length", "store_matrix",
     "treatment", "control"])

egs = 2290813547.4  # this is the effective genome size used by the original sicer for hg19


@pytest.fixture(scope="session")
def args_200_fast():
    return MockNamespace(25, "hg19", False, 200, 150, False, 3, 1, egs, False,
                         ["examples/test.bed"], ["examples/control.bed"])


@pytest.fixture(scope="session")
def args_200():
    return MockNamespace(1, "hg19", False, 200, 150, False, 3, 0.05, egs,
                         False, ["examples/test.bed"],
                         ["examples/control.bed"])


@pytest.fixture(scope="session")
def args_50():
    return MockNamespace(1, "hg19", False, 50, 150, False, 3, 0.05, egs, False,
                         ["examples/test.bed"], ["examples/control.bed"])


@pytest.fixture(scope="session")
def expected_result_example_input():
    return pd.read_table("examples/expected_results.csv", sep=" ", skiprows=1)



def remove_out_of_bounds_bins(df, chromosome_size):
    """Remove all reads that were shifted outside of the genome endpoints."""

    # The dataframe is empty and contains no bins out of bounds
    if "Bin" not in df:
        return df

    df = df.drop(df[df.Bin > chromosome_size].index)

    return df.drop(df[df.Bin < 0].index)



from itertools import count
from scipy.stats import poisson

from epic.config.constants import WINDOW_P_VALUE


def compute_enriched_threshold(average_window_readcount):
    """
    Computes the minimum number of tags required in window for an island to be enriched.
    """

    current_threshold, survival_function = 0, 1
    for current_threshold in count(start=0, step=1):
        survival_function -= poisson.pmf(current_threshold,
                                         average_window_readcount)
        if survival_function <= WINDOW_P_VALUE:
            break

    island_enriched_threshold = current_threshold + 1

    return island_enriched_threshold


def compute_gap_factor(island_enriched_threshold, gap_intervals_allowed,
                       poisson_distribution_parameter):

    max_gap_score = 1
    gap_factor = single_gap_factor(island_enriched_threshold,
                                   poisson_distribution_parameter)
    max_gap_score += sum([pow(gap_factor, i)
                          for i in range(1, gap_intervals_allowed + 1)])
    return max_gap_score


def single_gap_factor(island_enriched_threshold,
                      poisson_distribution_parameter):

    poisson_scores = [poisson.pmf(i, poisson_distribution_parameter)
                      for i in range(island_enriched_threshold)]
    return sum(poisson_scores)


def compute_boundary(island_enriched_threshold, gap_intervals_allowed,
                     average):

    single_gap = single_gap_factor(island_enriched_threshold, average)
    single_boundary_score = pow(single_gap, gap_intervals_allowed + 1)
    start_and_end_score = single_boundary_score * single_boundary_score

    return start_and_end_score



import logging
from natsort import natsorted

from joblib import Parallel, delayed

try:
    from functools import lru_cache  # noqa
except ImportError:
    from functools32 import lru_cache  # noqa


def _merge_chip_and_input(chip_df, input_df):

    chip_df_nb_bins = len(chip_df)
    merged_df = chip_df.merge(input_df,
                              how="left",
                              on=["Chromosome", "Bin"],
                              suffixes=[" ChIP", " Input"])
    merged_df = merged_df[["Chromosome", "Bin", "Count ChIP", "Count Input"]]
    merged_df.columns = ["Chromosome", "Bin", "ChIP", "Input"]

    merged_df = merged_df.fillna(0)

    assert len(merged_df) == chip_df_nb_bins

    return merged_df


def merge_chip_and_input(chip_dfs, input_dfs, nb_cpu):

    # should be same length, since missing chromos get empty df
    assert len(chip_dfs) == len(input_dfs)

    logging.info("Merging ChIP and Input data.")
    merged_chromosome_dfs = Parallel(n_jobs=nb_cpu)(
        delayed(_merge_chip_and_input)(chip_df, input_df)
        for chip_df, input_df in zip(chip_dfs, input_dfs))
    return merged_chromosome_dfs


def get_total_number_of_reads(dfs):
    return sum([df.Count.sum() for df in dfs])


def ensure_same_chromosomes_in_list(sample1_dfs, sample2_dfs):

    d1 = create_chromsome_df_map(sample1_dfs)
    d2 = create_chromsome_df_map(sample2_dfs)

    d1, d2 = fill_missing_chromosomes(d1, d2)

    sample1_dfs = [v for (k, v) in natsorted(d1.items())]
    sample2_dfs = [v for (k, v) in natsorted(d2.items())]

    return sample1_dfs, sample2_dfs


def create_chromsome_df_map(dfs):

    sample_dict = {}
    for df in dfs:
        chromosome = df.head(1).Chromosome.values[0]
        sample_dict[chromosome] = df

    return sample_dict


def fill_missing_chromosomes(d1, d2):

    all_chromosomomes = set(d1.keys()).union(d2.keys())

    d1_missing = set(d2.keys()).difference(d1.keys())
    d2_missing = set(d1.keys()).difference(d2.keys())

    for chromosome in d1_missing:
        d1[chromosome] = pd.DataFrame(columns=["Chromosome", "Bin"])

    for chromosome in d2_missing:
        d2[chromosome] = pd.DataFrame(columns=["Chromosome", "Bin"])

    return d1, d2


def merge_same_files(sample1_dfs, sample2_dfs, nb_cpu):

    # if one list is missing a chromosome, we might pair up the wrong dataframes
    # therefore creating dicts beforehand to ensure they are paired up properly
    sample1_dfs, sample2_dfs = ensure_same_chromosomes_in_list(sample1_dfs,
                                                               sample2_dfs)

    logging.info("Merging same class data.")
    merged_chromosome_dfs = Parallel(n_jobs=nb_cpu)(
        delayed(_merge_same_files)(sample1_df, sample2_df)
        for sample1_df, sample2_df in zip(sample1_dfs, sample2_dfs))

    return merged_chromosome_dfs


def _merge_same_files(sample1_df, sample2_df):

    merged_df = sample1_df.merge(sample2_df,
                                 how="outer",
                                 on=["Chromosome", "Bin"])

    return merged_df.fillna(0)



from operator import neg as minus
from math import log
from epic.statistics.compute_poisson import _poisson
import pandas as pd
from functools import partial
from joblib import Parallel, delayed
import logging


def count_to_pvalue(merged_dfs, island_enriched_threshold,
                    average_window_readcount, nb_cpu):

    logging.info("Giving bins poisson score.")
    parallel_count = partial(_count_to_pvalue, island_enriched_threshold,
                             average_window_readcount)
    return Parallel(n_jobs=nb_cpu)(delayed(parallel_count)(df)
                                   for df in merged_dfs)


def _count_to_pvalue(island_enriched_threshold, average_window_readcount, df):

    df = df.loc[df["ChIP"] >= island_enriched_threshold]
    scores = df["ChIP"].apply(
        lambda count: _poisson(count, average_window_readcount) + 0.000001)
    df.insert(len(df.columns), "Score", scores)

    pd.options.mode.chained_assignment = None

    df["Score"] = df["Score"].apply(log)
    df["Score"] = df["Score"].apply(minus)

    pd.options.mode.chained_assignment = "warn"

    return df



import pkg_resources
import logging

from epic.config import logging_settings
from epic.utils.find_readlength import (find_readlength,
                                        get_closest_readlength)

__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"


def get_genome_size_file(genome):
    return pkg_resources.resource_filename(
        "epic", "scripts/chromsizes/{}.chromsizes".format(genome))


def create_genome_size_dict(genome):
    """Creates genome size dict from string containing data."""

    size_file = get_genome_size_file(genome)
    size_lines = open(size_file).readlines()

    size_dict = {}
    for line in size_lines:
        genome, length = line.split()
        size_dict[genome] = int(length)

    return size_dict


def get_effective_genome_length(genome, read_length):

    egf = pkg_resources.resource_string(
        "epic", "scripts/effective_sizes/{}_{}.txt".format(
            genome, read_length)).split()[-1].decode()

    genome_length = sum(create_genome_size_dict(genome).values())

    logging.info("Using an effective genome fraction of {}.".format(egf))

    return float(egf) * genome_length



from epic.config.constants import E_VALUE, BIN_SIZE


def generate_cumulative_dist(island_expectations_d, total_length):
    """
    Generate cumulative distribution: a list of tuples (bins, hist).
    """

    cumulative = [0] * (total_length + 1)
    partial_sum = 0.0

    island_expectations = []
    for i in range(len(cumulative)):
        if i in island_expectations_d:
            island_expectations.append(island_expectations_d[i])
        else:
            island_expectations.append(0)

    for index in range(1, len(island_expectations) + 1):
        complimentary = len(island_expectations) - index
        partial_sum += island_expectations[complimentary]
        cumulative[complimentary] = partial_sum

    # move to function call
    for index in range(len(cumulative)):
        if cumulative[index] <= E_VALUE:
            score_threshold = index * BIN_SIZE
            break

    return score_threshold



from __future__ import print_function, division

import sys
import atexit
from subprocess import call, check_output
import os
from os.path import basename

from pyfaidx import Fasta


def effective_genome_size(fasta, read_length, nb_cores, tmpdir=None):
    """Compute effective genome size for genome."""

    idx = Fasta(fasta)

    genome_length = sum([len(c) for c in idx])

    print("File analyzed: ", fasta)
    print("Genome length: ", genome_length)

    chromosomes = ", ".join([c.name for c in idx])

    if "_" in chromosomes:
        print("Warning. The following chromosomes are part of your genome:\n",
              chromosomes.replace(">", "") + "\n",
              file=sys.stderr)
        print(
            "You probably want to remove all chromosomes in your fasta containing '_' for the effective genome size computation to be accurate.",
            file=sys.stderr)

    if tmpdir is None:
        try:
            tmpdir = os.environ['TMPDIR']
        except KeyError:
            tmpdir = '/tmp'
    output_file = os.path.join(tmpdir, '{1}.jf'.format(read_length, basename(fasta)))
    atexit.register(
        lambda: call("rm {output_file}".format(output_file=output_file), shell=True))

    call(
        "jellyfish count -t {nb_cores} -m {read_length} -s {genome_length} -L 1 -U 1 --out-counter-len 1 --counter-len 1 {fasta} -o {output_file}".format(
            **vars()),
        shell=True)

    stats = check_output("jellyfish stats {output_file}".format(
        output_file=output_file),
                         shell=True)

    unique_kmers = int(stats.split()[1])

    effective_genome_size = unique_kmers / genome_length

    print("Number unique {read_length}-mers: ".format(read_length=read_length),
          unique_kmers)
    print("Effective genome size: ", effective_genome_size)



import pytest
# from collections import namedtuple

from io import StringIO
import pandas as pd
from numpy import allclose, array_equal, int32

from epic.windows.count.count_reads_in_windows import (
    count_reads_in_windows, _count_reads_in_windows_paired_end)


@pytest.mark.integration
def test_count_reads_in_windows(expected_result, input_bed_file, args_200):

    actual_result = count_reads_in_windows(input_bed_file, args_200)

    actual_result = pd.concat(actual_result)

    print(actual_result.values, "actual")
    print(expected_result.values, "expected")

    assert array_equal(actual_result, expected_result)


@pytest.fixture
def input_bed_file(tmpdir):
    bed_file = tmpdir.join("read_file.bed")
    bed_file.write("""chr1	39036822	39036847	U0	0	+
chr1	39036870	39036900	U0	0	-
chr1	73781101	73781126	U0	0	+
chr1	90059861	90059886	U0	0	-
chr3	55648137	55648162	U0	0	+
chr7	20246668	20246693	U0	0	+
chr7	91135110	91135135	U0	0	+
chr13	100938475	100938500	U0	0	+
chr13	115816130	115816155	U0	0	+
chr19	43528773	43528798	U0	0	+
chr19	47109000	47109025	U0	0	-""")
    return str(bed_file)


@pytest.fixture
def expected_result(input_bed_file):

    df = pd.read_table(
        StringIO(u"""Count Chromosome Bin
    2 chr1 39036800
    1 chr1 73781000
    1 chr1 90059800
    1 chr3 55648200
    1 chr7 20246600
    1 chr7 91135000
    1 chr13 100938400
    1 chr19 43528800
    1 chr19 47108800"""),
        sep=r"\s+",
        dtype={"Count": int32,
               "Bin": int32})
    df.columns = [input_bed_file, "Chromosome", "Bin"]
    return df


@pytest.fixture
def paired_end(tmpdir):

    bed_file = tmpdir.join("paired_end_file.bed")
    bed_file.write(
        """chr1	33076	33165	chr1	33076	33165	2PJ3LS1:183:C5RR7ACXX:7:1307:11171:75017	0	+	-
chr1	33076	33165	chr1	33076	33165	2PJ3LS1:183:C5RR7ACXX:8:1115:18054:85095	0	+	-
chr1	33660	33744	chr1	33660	33744	2PJ3LS1:183:C5RR7ACXX:2:2201:20777:74838	0	+	-
chr1	33660	33744	chr1	33660	33744	2PJ3LS1:183:C5RR7ACXX:3:1213:9620:83082	0	+	-
chr1	38474	38571	chr1	38474	38571	2PJ3LS1:183:C5RR7ACXX:4:1216:1576:83235	0	+	-
chr1	41985	42043	chr1	42040	42131	2PJ3LS1:183:C5RR7ACXX:3:1214:11778:16074	1	+	-
chr1	42223	42324	chr1	42285	42386	2PJ3LS1:183:C5RR7ACXX:3:1215:18097:73129	0	+	-""")
    return str(bed_file)


@pytest.fixture
def expected_result_pe():
    df = pd.read_table(
        StringIO(u"""Count Chromosome    Bin
1       chr1  33000
1       chr1  33600
1       chr1  38400
1       chr1  42000
1       chr1  42200"""),
        sep="\s+",
        dtype={"Count": int32,
               "Bin": int32})
    return df


@pytest.mark.integration
def test_count_reads_in_windows_paired_end(paired_end, expected_result_pe):

    result = _count_reads_in_windows_paired_end(paired_end, False, 249250621,
                                                "chr1")

    print(result.dtypes)
    print(expected_result_pe)
    print(expected_result_pe.dtypes)
    assert result.equals(expected_result_pe)
